@@ -26,7 +26,7 @@ L{SQLObjectBase} is the central point of compatibility.
 import re
 
 from storm.properties import (
-    Unicode, Chars, Int, Bool, Float, DateTime, Date, TimeDelta)
+    RawStr, Int, Bool, Float, DateTime, Date, TimeDelta)
 from storm.references import Reference, ReferenceSet
 from storm.properties import SimpleProperty, PropertyPublisherMeta
 from storm.variables import Variable
@@ -96,7 +96,7 @@ class SQLObjectStyle(object):
         if name.startswith('_'):
             return name[1:]
         return name
-    
+
     def _mixed_to_under_sub(self, match):
         m = match.group(0).lower()
         if len(m) > 1:
@@ -115,7 +115,7 @@ class SQLObjectStyle(object):
     @staticmethod
     def _capword(s):
         return s[0].upper() + s[1:]
-    
+
     @staticmethod
     def _lowerword(s):
         return s[0].lower() + s[1:]
@@ -182,7 +182,7 @@ class SQLObjectMeta(PropertyPublisherMeta):
 
 
         id_type = dict.get("_idType", int)
-        id_cls = {int: Int, str: Chars, unicode: AutoUnicode}[id_type]
+        id_cls = {int: Int, str: RawStr, unicode: AutoUnicode}[id_type]
         dict[id_name] = id_cls(primary=True)
 
         # Notice that obj is the class since this is the metaclass.
@@ -458,8 +458,7 @@ class PropertyAdapter(object):
 class AutoUnicodeVariable(Variable):
     """Unlike UnicodeVariable, this will try to convert str to unicode."""
 
-    @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(self, value, from_db):
         if not isinstance(value, basestring):
             raise TypeError("Expected basestring, found %s" % repr(type(value)))
         return unicode(value)
