@@ -21,6 +21,7 @@
 from datetime import datetime, date, time, timedelta
 from decimal import Decimal
 import cPickle as pickle
+import re
 
 from storm.exceptions import NoneError
 from storm import Undef
@@ -33,7 +34,8 @@ __all__ = [
     "BoolVariable",
     "IntVariable",
     "FloatVariable",
-    "CharsVariable",
+    "DecimalVariable",
+    "RawStrVariable",
     "UnicodeVariable",
     "DateTimeVariable",
     "DateVariable",
@@ -45,11 +47,30 @@ __all__ = [
 ]
 
 
+class LazyValue(object):
+    """Marker to be used as a base class on lazily evaluated values."""
+
+
+def raise_none_error(column):
+    if not column:
+        raise NoneError("None isn't acceptable as a value")
+    else:
+        from storm.expr import compile, CompileError
+        name = column.name
+        if column.table is not Undef:
+            try:
+                table = compile(column.table)
+                name = "%s.%s" % (table, name)
+            except CompileError:
+                pass
+        raise NoneError("None isn't acceptable as a value for %s" % name)
+
+
 def VariableFactory(cls, **old_kwargs):
     """Build cls with kwargs of constructor updated by kwargs of call.
 
     This is really an implementation of partial/curry functions, and
-    should be replaced by 'partial' once 2.5 is in use.
+    is replaced by 'partial' when 2.5+ is in use.
     """
     def variable_factory(**new_kwargs):
         kwargs = old_kwargs.copy()
@@ -57,12 +78,24 @@ def VariableFactory(cls, **old_kwargs):
         return cls(**kwargs)
     return variable_factory
 
+try:
+    from functools import partial as VariableFactory
+except ImportError:
+    pass
+
 
 class Variable(object):
+    """Basic representation of a database value in Python.
+
+    @type column: L{storm.expr.Column}
+    @ivar column: The column this variable represents.
+    @type event: L{storm.event.EventSystem}
+    @ivar event: The event system on which to broadcast events. If
+        None, no events will be emitted.
+    """
 
     _value = Undef
     _lazy_value = Undef
-    _saved_state = Undef
     _checkpoint_state = Undef
     _allow_none = True
 
@@ -71,6 +104,24 @@ class Variable(object):
 
     def __init__(self, value=Undef, value_factory=Undef, from_db=False,
                  allow_none=True, column=None, event=None):
+        """
+        @param value: The initial value of this variable. The default
+            behavior is for the value to stay undefined until it is
+            set with L{set}.
+        @param value_factory: If specified, this will immediately be
+            called to get the initial value.
+        @param from_db: A boolean value indicating where the initial
+            value comes from, if C{value} or C{value_factory} are
+            specified.
+        @param allow_none: A boolean indicating whether None should be
+            allowed to be set as the value of this variable.
+        @type column: L{storm.expr.Column}
+        @param column: The column that this variable represents. It's
+            used for reporting better error messages.
+        @type event: L{EventSystem}
+        @param event: The event system to broadcast messages with. If
+            not specified, then no events will be broadcast.
+        """
         if not allow_none:
             self._allow_none = False
         if value is not Undef:
@@ -80,114 +131,185 @@ class Variable(object):
         self.column = column
         self.event = event
 
-    @staticmethod
-    def _parse_get(value, to_db):
-        return value
-
-    @staticmethod
-    def _parse_set(value, from_db):
-        return value
-
     def get_lazy(self, default=None):
+        """Get the current L{LazyValue} without resolving its value.
+
+        @param default: If no L{LazyValue} was previously specified,
+            return this value. Defaults to None.
+        """
         if self._lazy_value is Undef:
             return default
         return self._lazy_value
 
     def get(self, default=None, to_db=False):
+        """Get the value, resolving it from a L{LazyValue} if necessary.
+
+        If the current value is an instance of L{LazyValue}, then the
+        C{resolve-lazy-value} event will be emitted, to give third
+        parties the chance to resolve the lazy value to a real value.
+
+        @param default: Returned if no value has been set.
+        @param to_db: A boolean flag indicating whether this value is
+            destined for the database.
+        """
         if self._lazy_value is not Undef and self.event is not None:
             self.event.emit("resolve-lazy-value", self, self._lazy_value)
         value = self._value
         if value is Undef:
             return default
         if value is None:
-            return value
-        return self._parse_get(value, to_db)
+            return None
+        return self.parse_get(value, to_db)
 
     def set(self, value, from_db=False):
+        """Set a new value.
+
+        Generally this will be called when an attribute was set in
+        Python, or data is being loaded from the database.
+
+        If the value is different from the previous value (or it is a
+        L{LazyValue}), then the C{changed} event will be emitted.
+
+        @param value: The value to set. If this is an instance of
+            L{LazyValue}, then later calls to L{get} will try to
+            resolve the value.
+        @param from_db: A boolean indicating whether this value has
+            come from the database.
+        """
+        # FASTPATH This method is part of the fast path.  Be careful when
+        #          changing it (try to profile any changes).
+
         if isinstance(value, LazyValue):
             self._lazy_value = value
             new_value = Undef
         else:
-            if self._lazy_value is not Undef:
-                del self._lazy_value
+            self._lazy_value = Undef
             if value is None:
                 if self._allow_none is False:
-                    raise self._get_none_error()
+                    raise_none_error(self.column)
                 new_value = None
             else:
-                new_value = self._parse_set(value, from_db)
+                new_value = self.parse_set(value, from_db)
                 if from_db:
                     # Prepare it for being used by the hook below.
-                    value = self._parse_get(new_value, False)
+                    value = self.parse_get(new_value, False)
         old_value = self._value
         self._value = new_value
         if (self.event is not None and
             (self._lazy_value is not Undef or new_value != old_value)):
             if old_value is not None and old_value is not Undef:
-                old_value = self._parse_get(old_value, False)
+                old_value = self.parse_get(old_value, False)
             self.event.emit("changed", self, old_value, value, from_db)
 
     def delete(self):
+        """Delete the internal value.
+
+        If there was a value set, then emit the C{changed} event.
+        """
         old_value = self._value
-        if old_value != Undef:
+        if old_value is not Undef:
             self._value = Undef
-            if self.event:
+            if self.event is not None:
                 if old_value is not None and old_value is not Undef:
-                    old_value = self._parse_get(old_value, False)
+                    old_value = self.parse_get(old_value, False)
                 self.event.emit("changed", self, old_value, Undef, False)
 
     def is_defined(self):
+        """Check whether there is currently a value.
+
+        @return: boolean indicating whether there is currently a value
+            for this variable. Note that if a L{LazyValue} was
+            previously set, this returns False; it only returns True if
+            there is currently a real value set.
+        """
         return self._value is not Undef
 
     def has_changed(self):
+        """Check whether the value has changed.
+
+        @return: boolean indicating whether the value has changed
+            since the last call to L{checkpoint}.
+        """
         return (self._lazy_value is not Undef or
                 self.get_state() != self._checkpoint_state)
 
     def get_state(self):
+        """Get the internal state of this object.
+
+        @return: A value which can later be passed to L{set_state}.
+        """
         return (self._lazy_value, self._value)
 
     def set_state(self, state):
+        """Set the internal state of this object.
+
+        @param state: A result from a previous call to
+            L{get_state}. The internal state of this variable will be set
+            to the state of the variable which get_state was called on.
+        """
         self._lazy_value, self._value = state
 
     def checkpoint(self):
+        """"Checkpoint" the internal state.
+
+        See L{has_changed}.
+        """
         self._checkpoint_state = self.get_state()
 
     def copy(self):
-        variable = object.__new__(self.__class__)
+        """Make a new copy of this Variable with the same internal state."""
+        variable = self.__class__.__new__(self.__class__)
         variable.set_state(self.get_state())
         return variable
 
     def __eq__(self, other):
+        """Equality based on current value, not identity."""
         return (self.__class__ is other.__class__ and
                 self._value == other._value)
 
     def __hash__(self):
+        """Hash based on current value, not identity."""
         return hash(self._value)
 
-    def _get_none_error(self):
-        if not self.column:
-            return NoneError("None isn't acceptable as a value")
-        else:
-            from storm.expr import compile, CompileError
-            column = self.column.name
-            if self.column.table is not Undef:
-                try:
-                    table, parameters = compile(self.column.table)
-                    column = "%s.%s" % (table, column)
-                except CompileError:
-                    pass
-            return NoneError("None isn't acceptable as a value for %s"
-                             % column)
+    def parse_get(self, value, to_db):
+        """Convert the internal value to an external value.
+
+        Get a representation of this value either for Python or for
+        the database. This method is only intended to be overridden
+        in subclasses, not called from external code.
+
+        @param value: The value to be converted.
+        @param to_db: Whether or not this value is destined for the
+            database.
+        """
+        return value
+
+    def parse_set(self, value, from_db):
+        """Convert an external value to an internal value.
+
+        A value is being set either from Python code or from the
+        database. Parse it into its internal representation.  This
+        method is only intended to be overridden in subclasses, not
+        called from external code.
+
+        @param value: The value, either from Python code setting an
+            attribute or from a column in a database.
+        @param from_db: A boolean flag indicating whether this value
+            is from the database.
+        """
+        return value
 
 
-class LazyValue(object):
-    """Marker to be used as a base class on lazily evaluated values."""
+try:
+    from storm.cextensions import Variable
+except ImportError, e:
+    if "cextensions" not in str(e):
+        raise
 
 
 class BoolVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(self, value, from_db):
         if not isinstance(value, (int, long, float, Decimal)):
             raise TypeError("Expected bool, found %r: %r"
                             % (type(value), value))
@@ -196,8 +318,7 @@ class BoolVariable(Variable):
 
 class IntVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(self, value, from_db):
         if not isinstance(value, (int, long, float, Decimal)):
             raise TypeError("Expected int, found %r: %r"
                             % (type(value), value))
@@ -206,18 +327,35 @@ class IntVariable(Variable):
 
 class FloatVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(self, value, from_db):
         if not isinstance(value, (int, long, float, Decimal)):
             raise TypeError("Expected float, found %r: %r"
                             % (type(value), value))
         return float(value)
 
 
-class CharsVariable(Variable):
+class DecimalVariable(Variable):
 
     @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(value, from_db):
+        if (from_db and isinstance(value, basestring) or
+            isinstance(value, (int, long))):
+            value = Decimal(value)
+        elif not isinstance(value, Decimal):
+            raise TypeError("Expected Decimal, found %r: %r"
+                            % (type(value), value))
+        return value
+
+    @staticmethod
+    def parse_get(value, to_db):
+        if to_db:
+            return str(value)
+        return value
+
+
+class RawStrVariable(Variable):
+
+    def parse_set(self, value, from_db):
         if isinstance(value, buffer):
             value = str(value)
         elif not isinstance(value, str):
@@ -228,8 +366,7 @@ class CharsVariable(Variable):
 
 class UnicodeVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, from_db):
+    def parse_set(self, value, from_db):
         if not isinstance(value, unicode):
             raise TypeError("Expected unicode, found %r: %r"
                             % (type(value), value))
@@ -242,8 +379,8 @@ class DateTimeVariable(Variable):
         self._tzinfo = kwargs.pop("tzinfo", None)
         super(DateTimeVariable, self).__init__(*args, **kwargs)
 
-    def _parse_set(self, value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             if isinstance(value, datetime):
                 pass
             elif isinstance(value, (str, unicode)):
@@ -271,9 +408,8 @@ class DateTimeVariable(Variable):
 
 class DateVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             if value is None:
                 return None
             if isinstance(value, date):
@@ -293,9 +429,8 @@ class DateVariable(Variable):
 
 class TimeVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             # XXX Can None ever get here, considering that set() checks for it?
             if value is None:
                 return None
@@ -316,9 +451,8 @@ class TimeVariable(Variable):
 
 class TimeDeltaVariable(Variable):
 
-    @staticmethod
-    def _parse_set(value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             # XXX Can None ever get here, considering that set() checks for it?
             if value is None:
                 return None
@@ -326,15 +460,7 @@ class TimeDeltaVariable(Variable):
                 return value
             if not isinstance(value, (str, unicode)):
                 raise TypeError("Expected timedelta, found %s" % repr(value))
-            (years, months, days,
-             hours, minutes, seconds, microseconds) = _parse_interval(value)
-            if years != 0:
-                raise ValueError("Can not handle year intervals")
-            if months != 0:
-                raise ValueError("Can not handle month intervals")
-            return timedelta(days=days, hours=hours,
-                             minutes=minutes, seconds=seconds,
-                             microseconds=microseconds)
+            return _parse_interval(value)
         else:
             if not isinstance(value, timedelta):
                 raise TypeError("Expected timedelta, found %s" % repr(value))
@@ -348,16 +474,16 @@ class EnumVariable(Variable):
         self._set_map = set_map
         Variable.__init__(self, *args, **kwargs)
 
-    def _parse_set(self, value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             return value
         try:
             return self._set_map[value]
         except KeyError:
             raise ValueError("Invalid enum value: %s" % repr(value))
 
-    def _parse_get(self, value, db):
-        if db:
+    def parse_get(self, value, to_db):
+        if to_db:
             return value
         try:
             return self._get_map[value]
@@ -377,18 +503,16 @@ class PickleVariable(Variable):
         if self.get_state() != self._checkpoint_state:
             self.event.emit("changed", self, None, self._value, False)
 
-    @staticmethod
-    def _parse_set(value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             if isinstance(value, buffer):
                 value = str(value)
             return pickle.loads(value)
         else:
             return value
 
-    @staticmethod
-    def _parse_get(value, db):
-        if db:
+    def parse_get(self, value, to_db):
+        if to_db:
             return pickle.dumps(value, -1)
         else:
             return value
@@ -420,17 +544,19 @@ class ListVariable(Variable):
         if self.get_state() != self._checkpoint_state:
             self.event.emit("changed", self, None, self._value, False)
 
-    def _parse_set(self, value, db):
-        if db:
+    def parse_set(self, value, from_db):
+        if from_db:
             item_factory = self._item_factory
-            return [item_factory(value=val, from_db=db).get() for val in value]
+            return [item_factory(value=val, from_db=from_db).get()
+                    for val in value]
         else:
             return value
 
-    def _parse_get(self, value, db):
-        if db:
+    def parse_get(self, value, to_db):
+        if to_db:
             item_factory = self._item_factory
-            return [item_factory(value=val, from_db=db) for val in value]
+            # XXX This from_db=to_db is dubious. What to do here?
+            return [item_factory(value=val, from_db=to_db) for val in value]
         else:
             return value
 
@@ -447,9 +573,14 @@ class ListVariable(Variable):
 
 def _parse_time(time_str):
     # TODO Add support for timezones.
-    if ":" not in time_str:
+    colons = time_str.count(":")
+    if not 1 <= colons <= 2:
         raise ValueError("Unknown time format: %r" % time_str)
-    hour, minute, second = time_str.split(":")
+    if colons == 2:
+        hour, minute, second = time_str.split(":")
+    else:
+        hour, minute = time_str.split(":")
+        second = "0"
     if "." in second:
         second, microsecond = second.split(".")
         second = int(second)
@@ -463,61 +594,53 @@ def _parse_date(date_str):
     year, month, day = date_str.split("-")
     return int(year), int(month), int(day)
 
-# XXX: 20070208 jamesh
-# The function below comes from psycopgda.adapter, which is licensed
-# under the ZPL.  Depending on Storm licensing, we may need to replace
-# it.
-def _parse_interval(interval_str):
-    """Parses PostgreSQL interval notation and returns a tuple (years, months,
-    days, hours, minutes, seconds).
 
-    Values accepted:
-        interval  ::= date
-                   |  time
-                   |  date time
-        date      ::= date_comp
-                   |  date date_comp
-        date_comp ::= 1 'day'
-                   |  number 'days'
-                   |  1 'month'
-                   |  1 'mon'
-                   |  number 'months'
-                   |  number 'mons'
-                   |  1 'year'
-                   |  number 'years'
-        time      ::= number ':' number
-                   |  number ':' number ':' number
-                   |  number ':' number ':' number '.' fraction
-    """
-    years = months = days = 0
-    hours = minutes = seconds = microseconds = 0
-    elements = interval_str.split()
-    # Tests with 7.4.6 on Ubuntu 5.4 interval output returns 'mon' and 'mons'
-    # and not 'month' or 'months' as expected. I've fixed this and left
-    # the original matches there too in case this is dependant on
-    # OS or PostgreSQL release.
-    for i in range(0, len(elements) - 1, 2):
-        count, unit = elements[i:i+2]
-        if unit.endswith(','):
-            unit = unit[:-1]
-        if unit == 'day' and count == '1':
-            days += 1
-        elif unit == 'days':
-            days += int(count)
-        elif unit == 'month' and count == '1':
-            months += 1
-        elif unit == 'mon' and count == '1':
-            months += 1
-        elif unit == 'months':
-            months += int(count)
-        elif unit == 'mons':
-            months += int(count)
-        elif unit == 'year' and count == '1':
-            years += 1
-        elif unit == 'years':
-            years += int(count)
+def _parse_interval_table():
+    table = {}
+    for units, delta in (
+        ("d day days", timedelta),
+        ("h hour hours", lambda x: timedelta(hours=x)),
+        ("m min minute minutes", lambda x: timedelta(minutes=x)),
+        ("s sec second seconds", lambda x: timedelta(seconds=x)),
+        ("ms millisecond milliseconds", lambda x: timedelta(milliseconds=x)),
+        ("microsecond microseconds", lambda x: timedelta(microseconds=x))
+        ):
+        for unit in units.split():
+            table[unit] = delta
+    return table
+
+_parse_interval_table = _parse_interval_table()
+
+_parse_interval_re = re.compile(r"[\s,]*"
+                                r"([-+]?(?:\d\d?:\d\d?(?::\d\d?)?(?:\.\d+)?"
+                                r"|\d+(?:\.\d+)?))"
+                                r"[\s,]*")
+
+def _parse_interval(interval):
+    result = timedelta(0)
+    value = None
+    for token in _parse_interval_re.split(interval):
+        if not token:
+            pass
+        elif ":" in token:
+            if value is not None:
+                result += timedelta(days=value)
+                value = None
+            h, m, s, ms = _parse_time(token)
+            result += timedelta(hours=h, minutes=m, seconds=s, microseconds=ms)
+        elif value is None:
+            try:
+                value = float(token)
+            except ValueError:
+                raise ValueError("Expected an interval value rather than "
+                                 "%r in interval %r" % (token, interval))
         else:
-            raise ValueError, 'unknown time interval %s %s' % (count, unit)
-    if len(elements) % 2 == 1:
-        hours, minutes, seconds, microseconds = _parse_time(elements[-1])
-    return (years, months, days, hours, minutes, seconds, microseconds)
+            unit = _parse_interval_table.get(token)
+            if unit is None:
+                raise ValueError("Unsupported interval unit %r in interval %r"
+                                 % (token, interval))
+            result += unit(value)
+            value = None
+    if value is not None:
+        result += timedelta(seconds=value)
+    return result
